import { BigintIsh, CurrencyAmount, Price, Token } from '@uniswap/sdk-core';
import JSBI from 'jsbi';
import { FeeAmount } from '../constants';
import { Tick, TickConstructorArgs } from './tick';
import { TickDataProvider } from './tickDataProvider';
import { ethers } from 'ethers';
interface SnapshotCumulativeInside {
    tickCumulativeInside: bigint;
    secondsPerLiquidityInsideX128: bigint;
    secondsInside: bigint;
}
interface ObserveResponse {
    tickCumulatives: bigint[];
    secondsPerLiquidityCumulativeX128s: bigint[];
}
interface TransactionOverrides {
    gasPrice?: BigInt;
    gasLimit?: BigInt;
    value?: BigInt;
    nonce?: BigInt;
}
interface Slot0Response {
    sqrtPriceX96: BigInt;
    tick: number;
    observationIndex: number;
    observationCardinality: number;
    observationCardinalityNext: number;
    feeProtocol: number;
    unlocked: boolean;
}
/**
 * Represents a V3 pool
 */
export declare class Pool {
    readonly token0: Token;
    readonly token1: Token;
    readonly fee: FeeAmount;
    get sqrtRatioX96(): JSBI;
    readonly _sqrtRatioX96: bigint;
    get liquidity(): JSBI;
    readonly _liquidity: bigint;
    readonly tickCurrent: number;
    get tickDataProvider(): TickDataProvider;
    private _tickDataProvider;
    private _token0Price?;
    private _token1Price?;
    readonly _provider?: ethers.providers.Provider;
    static getAddress(tokenA: Token, tokenB: Token, fee: FeeAmount, initCodeHashManualOverride?: string, factoryAddressOverride?: string): string;
    /**
     * Construct a pool
     * @param tokenA One of the tokens in the pool
     * @param tokenB The other token in the pool
     * @param fee The fee in hundredths of a bips of the input amount of every swap that is collected by the pool
     * @param sqrtRatioX96 The sqrt of the current ratio of amounts of token1 to token0
     * @param liquidity The current value of in range liquidity
     * @param tickCurrent The current tick of the pool
     * @param ticks The current state of the pool ticks or a data provider that can return tick data
     */
    constructor(tokenA: Token, tokenB: Token, fee: FeeAmount, sqrtRatioX96: BigintIsh, liquidity: BigintIsh, tickCurrent: number, _ticks?: TickDataProvider | (Tick | TickConstructorArgs)[], provider?: ethers.providers.Provider);
    /**
     * Returns true if the token is either token0 or token1
     * @param token The token to check
     * @returns True if token is either token0 or token
     */
    involvesToken(token: Token): boolean;
    /**
     * Returns the current mid price of the pool in terms of token0, i.e. the ratio of token1 over token0
     */
    get token0Price(): Price<Token, Token>;
    /**
     * Returns the current mid price of the pool in terms of token1, i.e. the ratio of token0 over token1
     */
    get token1Price(): Price<Token, Token>;
    /**
     * Return the price of the given token in terms of the other token in the pool.
     * @param token The token to return price of
     * @returns The price of the given token, in terms of the other.
     */
    priceOf(token: Token): Price<Token, Token>;
    /**
     * Returns the chain ID of the tokens in the pool.
     */
    get chainId(): number;
    /**
     * Given an input amount of a token, return the computed output amount, and a pool with state updated after the trade
     * @param inputAmount The input amount for which to quote the output amount
     * @param sqrtPriceLimitX96 The Q64.96 sqrt price limit
     * @returns The output amount and the pool with updated state
     */
    getOutputAmount(inputAmount: CurrencyAmount<Token>, sqrtPriceLimitX96?: bigint | JSBI): Promise<[CurrencyAmount<Token>, Pool]>;
    /**
     * Given a desired output amount of a token, return the computed input amount and a pool with state updated after the trade
     * @param outputAmount the output amount for which to quote the input amount
     * @param sqrtPriceLimitX96 The Q64.96 sqrt price limit. If zero for one, the price cannot be less than this value after the swap. If one for zero, the price cannot be greater than this value after the swap
     * @returns The input amount and the pool with updated state
     */
    getInputAmount(outputAmount: CurrencyAmount<Token>, sqrtPriceLimitX96?: bigint | JSBI): Promise<[CurrencyAmount<Token>, Pool]>;
    /**
     * Executes a swap
     * @param zeroForOne Whether the amount in is token0 or token1
     * @param amountSpecified The amount of the swap, which implicitly configures the swap as exact input (positive), or exact output (negative)
     * @param sqrtPriceLimitX96 The Q64.96 sqrt price limit. If zero for one, the price cannot be less than this value after the swap. If one for zero, the price cannot be greater than this value after the swap
     * @returns amountCalculated
     * @returns sqrtRatioX96
     * @returns liquidity
     * @returns tickCurrent
     */
    private swap;
    get tickSpacing(): number;
    initializeTicks(provider?: ethers.providers.Provider | undefined): Promise<void>;
    initializeTicksFromRpc(): Promise<void>;
    static rpcCreatePool(_signer: ethers.Signer, provider: ethers.providers.Provider, tokenA: string, tokenB: string, fee: FeeAmount, transactionOverrides?: TransactionOverrides, factoryAddress?: string): Promise<ethers.providers.TransactionResponse>;
    private static ethersTransactionOverrides;
    rpcContract(poolAddress?: string, signer?: ethers.Signer): Promise<ethers.Contract>;
    rpcPoolExists(poolAddress?: string, blockNum?: number): Promise<boolean>;
    rpcSlot0(poolAddress?: string, blockNum?: number): Promise<Slot0Response>;
    rpcSnapshotCumulativesInside(tickLower: number, tickUpper: number, poolAddress?: string, blockNum?: number): Promise<SnapshotCumulativeInside>;
    rpcObserve(secondsAgo: number[], poolAddress?: string, blockNum?: number): Promise<ObserveResponse>;
    rpcIncreaseObservationCardinalityNext(signer: ethers.Signer, observationCardinalityNext: number, poolAddress?: string, transactionOverrides?: TransactionOverrides): Promise<ethers.providers.TransactionResponse>;
    rpcCollect(signer: ethers.Signer, recipient: string, tickLower: number, tickUpper: number, amount0Requested: BigInt, amount1Requested: BigInt, poolAddress?: string, transactionOverrides?: TransactionOverrides): Promise<ethers.providers.TransactionResponse>;
    rpcBurn(signer: ethers.Signer, tickLower: number, tickUpper: number, amount: BigInt, poolAddress?: string, transactionOverrides?: TransactionOverrides): Promise<ethers.providers.TransactionResponse>;
}
export {};
