import { CurrencyAmount, Price, V3_CORE_FACTORY_ADDRESSES } from '@uniswap/sdk-core';
import JSBI from 'jsbi';
import invariant from 'tiny-invariant';
import { FACTORY_ADDRESS, TICK_SPACINGS } from '../constants';
import { Q192_BIGINT } from '../internalConstants';
import { computePoolAddress } from '../utils/computePoolAddress';
import { LiquidityMath } from '../utils/liquidityMath';
import { SwapMath } from '../utils/swapMath';
import { TickMath } from '../utils/tickMath';
import { NoTickDataProvider } from './tickDataProvider';
import { TickListDataProvider } from './tickListDataProvider';
import { bigIntFromBigintIsh } from 'src/utils/bigintIsh';
import { ethers } from 'ethers';
import { RPCTickDataProvider } from './rpcTickDataProvider';
import { abi as poolAbi } from '@uniswap/v3-core/artifacts/contracts/UniswapV3Pool.sol/UniswapV3Pool.json';
import { abi as factoryAbi } from '@uniswap/v3-core/artifacts/contracts/UniswapV3Factory.sol/UniswapV3Factory.json';
/**
 * By default, pools will not allow operations that require ticks.
 */
const NO_TICK_DATA_PROVIDER_DEFAULT = new NoTickDataProvider();
/**
 * Represents a V3 pool
 */
export class Pool {
    get sqrtRatioX96() {
        return JSBI.BigInt(this._sqrtRatioX96.toString(10));
    }
    get liquidity() {
        return JSBI.BigInt(this._liquidity.toString(10));
    }
    get tickDataProvider() {
        return this._tickDataProvider;
    }
    static getAddress(tokenA, tokenB, fee, initCodeHashManualOverride, factoryAddressOverride) {
        return computePoolAddress({
            factoryAddress: factoryAddressOverride ?? FACTORY_ADDRESS,
            fee,
            tokenA,
            tokenB,
            initCodeHashManualOverride,
        });
    }
    /**
     * Construct a pool
     * @param tokenA One of the tokens in the pool
     * @param tokenB The other token in the pool
     * @param fee The fee in hundredths of a bips of the input amount of every swap that is collected by the pool
     * @param sqrtRatioX96 The sqrt of the current ratio of amounts of token1 to token0
     * @param liquidity The current value of in range liquidity
     * @param tickCurrent The current tick of the pool
     * @param ticks The current state of the pool ticks or a data provider that can return tick data
     */
    constructor(tokenA, tokenB, fee, sqrtRatioX96, liquidity, tickCurrent, _ticks, provider) {
        invariant(Number.isInteger(fee) && fee < 1000000, 'FEE');
        const ticks = _ticks || NO_TICK_DATA_PROVIDER_DEFAULT;
        this._provider = provider;
        const tickCurrentSqrtRatioX96 = TickMath.getSqrtRatioAtTickBigInt(tickCurrent);
        const nextTickSqrtRatioX96 = TickMath.getSqrtRatioAtTickBigInt(tickCurrent + 1);
        invariant(bigIntFromBigintIsh(sqrtRatioX96) >= tickCurrentSqrtRatioX96 &&
            bigIntFromBigintIsh(sqrtRatioX96) <= nextTickSqrtRatioX96, 'PRICE_BOUNDS');
        [this.token0, this.token1] = tokenA.sortsBefore(tokenB) ? [tokenA, tokenB] : [tokenB, tokenA];
        this.fee = fee;
        this._sqrtRatioX96 = bigIntFromBigintIsh(sqrtRatioX96);
        this._liquidity = bigIntFromBigintIsh(liquidity);
        this.tickCurrent = tickCurrent;
        this._tickDataProvider = Array.isArray(ticks) ? new TickListDataProvider(ticks, TICK_SPACINGS[fee]) : ticks;
        if (this.tickDataProvider instanceof NoTickDataProvider && this._provider) {
            this._tickDataProvider = new RPCTickDataProvider(this._provider, Pool.getAddress(tokenA, tokenB, fee));
        }
    }
    /**
     * Returns true if the token is either token0 or token1
     * @param token The token to check
     * @returns True if token is either token0 or token
     */
    involvesToken(token) {
        return token.equals(this.token0) || token.equals(this.token1);
    }
    /**
     * Returns the current mid price of the pool in terms of token0, i.e. the ratio of token1 over token0
     */
    get token0Price() {
        return (this._token0Price ??
            (this._token0Price = new Price(this.token0, this.token1, Q192_BIGINT, this._sqrtRatioX96 * this._sqrtRatioX96)));
    }
    /**
     * Returns the current mid price of the pool in terms of token1, i.e. the ratio of token0 over token1
     */
    get token1Price() {
        return (this._token1Price ??
            (this._token1Price = new Price(this.token1, this.token0, this._sqrtRatioX96 * this._sqrtRatioX96, Q192_BIGINT)));
    }
    /**
     * Return the price of the given token in terms of the other token in the pool.
     * @param token The token to return price of
     * @returns The price of the given token, in terms of the other.
     */
    priceOf(token) {
        invariant(this.involvesToken(token), 'TOKEN');
        return token.equals(this.token0) ? this.token0Price : this.token1Price;
    }
    /**
     * Returns the chain ID of the tokens in the pool.
     */
    get chainId() {
        return this.token0.chainId;
    }
    /**
     * Given an input amount of a token, return the computed output amount, and a pool with state updated after the trade
     * @param inputAmount The input amount for which to quote the output amount
     * @param sqrtPriceLimitX96 The Q64.96 sqrt price limit
     * @returns The output amount and the pool with updated state
     */
    async getOutputAmount(inputAmount, sqrtPriceLimitX96) {
        invariant(this.involvesToken(inputAmount.currency), 'TOKEN');
        const zeroForOne = inputAmount.currency.equals(this.token0);
        const { amountCalculated: outputAmount, sqrtRatioX96, liquidity, tickCurrent, } = await this.swap(zeroForOne, inputAmount.quotient, sqrtPriceLimitX96);
        const outputToken = zeroForOne ? this.token1 : this.token0;
        const negation = BigInt(outputAmount.toString(10)) * -1n;
        return [
            CurrencyAmount.fromRawAmount(outputToken, negation),
            new Pool(this.token0, this.token1, this.fee, sqrtRatioX96, liquidity, tickCurrent, this.tickDataProvider),
        ];
    }
    /**
     * Given a desired output amount of a token, return the computed input amount and a pool with state updated after the trade
     * @param outputAmount the output amount for which to quote the input amount
     * @param sqrtPriceLimitX96 The Q64.96 sqrt price limit. If zero for one, the price cannot be less than this value after the swap. If one for zero, the price cannot be greater than this value after the swap
     * @returns The input amount and the pool with updated state
     */
    async getInputAmount(outputAmount, sqrtPriceLimitX96) {
        invariant(outputAmount.currency.isToken && this.involvesToken(outputAmount.currency), 'TOKEN');
        const zeroForOne = outputAmount.currency.equals(this.token1);
        const { amountCalculated: inputAmount, sqrtRatioX96, liquidity, tickCurrent, } = await this.swap(zeroForOne, outputAmount.quotientBigInt * -1n, sqrtPriceLimitX96);
        const inputToken = zeroForOne ? this.token0 : this.token1;
        return [
            CurrencyAmount.fromRawAmount(inputToken, inputAmount),
            new Pool(this.token0, this.token1, this.fee, sqrtRatioX96, liquidity, tickCurrent, this.tickDataProvider),
        ];
    }
    /**
     * Executes a swap
     * @param zeroForOne Whether the amount in is token0 or token1
     * @param amountSpecified The amount of the swap, which implicitly configures the swap as exact input (positive), or exact output (negative)
     * @param sqrtPriceLimitX96 The Q64.96 sqrt price limit. If zero for one, the price cannot be less than this value after the swap. If one for zero, the price cannot be greater than this value after the swap
     * @returns amountCalculated
     * @returns sqrtRatioX96
     * @returns liquidity
     * @returns tickCurrent
     */
    async swap(zeroForOne, _amountSpecified, _sqrtPriceLimitX96) {
        const amountSpecified = bigIntFromBigintIsh(_amountSpecified);
        const defaultSqrtPriceLimitX96 = zeroForOne
            ? TickMath.MIN_SQRT_RATIO_BIGINT + 1n
            : TickMath.MAX_SQRT_RATIO_BIGINT - 1n;
        const sqrtPriceLimitX96 = _sqrtPriceLimitX96 ? bigIntFromBigintIsh(_sqrtPriceLimitX96) : defaultSqrtPriceLimitX96;
        if (zeroForOne) {
            invariant(sqrtPriceLimitX96 > TickMath.MIN_SQRT_RATIO_BIGINT, 'RATIO_MIN');
            invariant(sqrtPriceLimitX96 < this._sqrtRatioX96, 'RATIO_CURRENT');
        }
        else {
            invariant(sqrtPriceLimitX96 < TickMath.MAX_SQRT_RATIO_BIGINT, 'RATIO_MAX');
            invariant(sqrtPriceLimitX96 > this._sqrtRatioX96, 'RATIO_CURRENT');
        }
        const exactInput = amountSpecified >= 0n;
        // keep track of swap state
        const state = {
            amountSpecifiedRemaining: amountSpecified,
            amountCalculated: 0n,
            sqrtPriceX96: this._sqrtRatioX96,
            tick: this.tickCurrent,
            liquidity: this._liquidity,
        };
        // start swap while loop
        while (state.amountSpecifiedRemaining !== 0n && state.sqrtPriceX96 !== sqrtPriceLimitX96) {
            let step = {};
            step.sqrtPriceStartX96 = state.sqrtPriceX96;
            [step.tickNext, step.initialized] = await this.tickDataProvider.nextInitializedTickWithinOneWord(state.tick, zeroForOne, this.tickSpacing);
            if (step.tickNext < TickMath.MIN_TICK) {
                step.tickNext = TickMath.MIN_TICK;
            }
            else if (step.tickNext > TickMath.MAX_TICK) {
                step.tickNext = TickMath.MAX_TICK;
            }
            step.sqrtPriceNextX96 = TickMath.getSqrtRatioAtTickBigInt(step.tickNext);
            [state.sqrtPriceX96, step.amountIn, step.amountOut, step.feeAmount] = SwapMath.computeSwapStep(state.sqrtPriceX96, (zeroForOne ? step.sqrtPriceNextX96 < sqrtPriceLimitX96 : step.sqrtPriceNextX96 > sqrtPriceLimitX96)
                ? sqrtPriceLimitX96
                : step.sqrtPriceNextX96, state.liquidity, state.amountSpecifiedRemaining, this.fee);
            if (exactInput) {
                state.amountSpecifiedRemaining = state.amountSpecifiedRemaining - (step.amountIn + step.feeAmount);
                state.amountCalculated = state.amountCalculated - step.amountOut;
            }
            else {
                state.amountSpecifiedRemaining = state.amountSpecifiedRemaining + step.amountOut;
                state.amountCalculated = state.amountCalculated + (step.amountIn + step.feeAmount);
            }
            // TODO
            if (state.sqrtPriceX96 === step.sqrtPriceNextX96) {
                // if the tick is initialized, run the tick transition
                if (step.initialized) {
                    let liquidityNet = bigIntFromBigintIsh((await this.tickDataProvider.getTick(step.tickNext)).liquidityNet);
                    // if we're moving leftward, we interpret liquidityNet as the opposite sign
                    // safe because liquidityNet cannot be type(int128).min
                    if (zeroForOne)
                        liquidityNet = liquidityNet * -1n;
                    state.liquidity = LiquidityMath.addDelta(state.liquidity, liquidityNet);
                }
                state.tick = zeroForOne ? step.tickNext - 1 : step.tickNext;
            }
            else if (state.sqrtPriceX96 !== step.sqrtPriceStartX96) {
                // updated comparison function
                // recompute unless we're on a lower tick boundary (i.e. already transitioned ticks), and haven't moved
                state.tick = TickMath.getTickAtSqrtRatio(state.sqrtPriceX96);
            }
        }
        if (typeof _amountSpecified === 'bigint') {
            return {
                amountCalculated: state.amountCalculated,
                sqrtRatioX96: state.sqrtPriceX96,
                liquidity: state.liquidity,
                tickCurrent: state.tick,
            };
        }
        else {
            return {
                amountCalculated: JSBI.BigInt(state.amountCalculated.toString(10)),
                sqrtRatioX96: JSBI.BigInt(state.sqrtPriceX96.toString(10)),
                liquidity: JSBI.BigInt(state.liquidity.toString(10)),
                tickCurrent: state.tick,
            };
        }
    }
    get tickSpacing() {
        return TICK_SPACINGS[this.fee];
    }
    // ---- RPC Functions - Fetch data from on-chain state ----
    async initializeTicks(provider) {
        if ((this.tickDataProvider instanceof NoTickDataProvider)) {
            invariant(provider !== undefined, 'Pool has no RPC connection and no Provider was provided.');
            this._tickDataProvider = new RPCTickDataProvider(provider, Pool.getAddress(this.token0, this.token1, this.fee));
        }
        await this.initializeTicksFromRpc();
        // If the TickDataProvider is neither a NoTickDataProvider nor an RPCTickDataProvider ticks are present
    }
    async initializeTicksFromRpc() {
        invariant(!(this.tickDataProvider instanceof NoTickDataProvider), 'Pool has no RPC connection');
        if (this.tickDataProvider instanceof RPCTickDataProvider) {
            await this.tickDataProvider.rpcFetchTicks();
        }
        // If the TickDataProvider is neither a NoTickDataProvider nor an RPCTickDataProvider ticks are present
    }
    static async rpcCreatePool(_signer, provider, tokenA, tokenB, fee, transactionOverrides, factoryAddress) {
        const signer = _signer.connect(provider);
        let factory;
        if (factoryAddress) {
            factory = factoryAddress;
        }
        else {
            const network = await provider.getNetwork();
            factory = V3_CORE_FACTORY_ADDRESSES[network.chainId];
        }
        const contract = new ethers.Contract(factory, factoryAbi, signer);
        const response = contract.createPool(tokenA, tokenB, fee, Pool.ethersTransactionOverrides(transactionOverrides));
        return response;
    }
    static ethersTransactionOverrides(transactionOverrides) {
        return {
            gasPrice: transactionOverrides?.gasPrice
                ? ethers.BigNumber.from(transactionOverrides.gasPrice.toString(10))
                : undefined,
            gasLimit: transactionOverrides?.gasLimit
                ? ethers.BigNumber.from(transactionOverrides.gasLimit.toString(10))
                : undefined,
            value: transactionOverrides?.value ? ethers.BigNumber.from(transactionOverrides.value.toString(10)) : undefined,
            nonce: transactionOverrides?.nonce ? ethers.BigNumber.from(transactionOverrides.nonce.toString(10)) : undefined,
        };
    }
    async rpcContract(poolAddress, signer) {
        invariant(this._provider, 'provider not initialized');
        const provider = signer ? signer.connect(this._provider) : this._provider;
        return new ethers.Contract(poolAddress || Pool.getAddress(this.token0, this.token1, this.fee), poolAbi, provider);
    }
    async rpcPoolExists(poolAddress, blockNum) {
        try {
            await this.rpcSlot0(poolAddress, blockNum);
            return true;
        }
        catch {
            return false;
        }
    }
    async rpcSlot0(poolAddress, blockNum) {
        invariant(this._provider, 'provider not initialized');
        const contract = await this.rpcContract(poolAddress);
        const response = await contract.slot0({ blockTag: blockNum || 'latest' });
        return {
            sqrtPriceX96: BigInt(response.sqrtPriceX96.toString()),
            tick: response.tick,
            observationIndex: response.observationIndex,
            observationCardinality: response.observationCardinality,
            observationCardinalityNext: response.observationCardinalityNext,
            feeProtocol: response.feeProtocol,
            unlocked: response.unlocked,
        };
    }
    async rpcSnapshotCumulativesInside(tickLower, tickUpper, poolAddress, blockNum) {
        invariant(this._provider, 'provider not initialized');
        const contract = await this.rpcContract(poolAddress);
        const response = await contract.snapshotCumulativesInside(tickLower, tickUpper, { blockTag: blockNum || 'latest' });
        return {
            secondsInside: BigInt(response.secondsInside.toString()),
            secondsPerLiquidityInsideX128: BigInt(response.secondsPerLiquidityInsideX128.toString()),
            tickCumulativeInside: BigInt(response.tickCumulativeInside.toString()),
        };
    }
    async rpcObserve(secondsAgo, poolAddress, blockNum) {
        invariant(this._provider, 'provider not initialized');
        const contract = await this.rpcContract(poolAddress);
        const response = await contract.observe(secondsAgo, { blockTag: blockNum || 'latest' });
        return {
            secondsPerLiquidityCumulativeX128s: response.secondsPerLiquidityCumulativeX128s.map((num) => BigInt(num.toString())),
            tickCumulatives: response.tickCumulatives.map((num) => BigInt(num.toString())),
        };
    }
    async rpcIncreaseObservationCardinalityNext(signer, observationCardinalityNext, poolAddress, transactionOverrides) {
        invariant(this._provider, 'provider not initialized');
        const contract = await this.rpcContract(poolAddress, signer);
        const response = await contract.increaseObservationCardinalityNext(observationCardinalityNext, Pool.ethersTransactionOverrides(transactionOverrides));
        return response;
    }
    async rpcCollect(signer, recipient, tickLower, tickUpper, amount0Requested, amount1Requested, poolAddress, transactionOverrides) {
        invariant(this._provider, 'provider not initialized');
        const contract = await this.rpcContract(poolAddress, signer);
        const response = await contract.collect(recipient, tickLower, tickUpper, amount0Requested, amount1Requested, Pool.ethersTransactionOverrides(transactionOverrides));
        return response;
    }
    async rpcBurn(signer, tickLower, tickUpper, amount, poolAddress, transactionOverrides) {
        invariant(this._provider, 'provider not initialized');
        const contract = await this.rpcContract(poolAddress, signer);
        const response = await contract.burn(tickLower, tickUpper, amount, Pool.ethersTransactionOverrides(transactionOverrides));
        return response;
    }
}
//# sourceMappingURL=pool.js.map